from __future__ import annotations
from typing import Any, List

from steam.steamid import SteamID

from .gamemodes import COMPETITIVE
from .map_source import MAP_GROUP

from .exceptions import (
    InvalidSlotSize,
    MultipleGames,
    InvalidTickrate,
    InvalidSteamID,
    InvalidStorageSize
)


VALID_TICKRATES = [
    64,
    85,
    100,
    102.4,
    128
]


class ServerSettings:
    __game = False

    def __init__(self, name: str = None, location: str = None,
                 custom_domain: str = None,
                 autostop: bool = None, autostop_minutes: int = None,
                 mysql: bool = None, scheduled_commands: List[str] = None,
                 user_data: str = None,
                 reboot_on_crash: bool = None,
                 max_disk_usage_gb: int = None,
                 manual_sort_order: int = None,
                 core_dump: bool = None) -> None:
        """Used to store settings on a server.

        Parameters
        ----------
        name : str, optional
            Name of server, by default None
        location : str, optional
            Location of server, by default None
        custom_domain : str, optional
            Custom domain name, by default None
        autostop : bool, optional
            If autostop is enabled, by default None
        autostop_minutes : int, optional
            How many minutes, by default None
        mysql : bool, optional
            If myself is enabled, by default None
        scheduled_commands : List[str], optional
            List of scheduled commands, by default None
        user_data : str, optional
            User meta data, by default None
        reboot_on_crash : bool, optional
            by default None
        max_disk_usage_gb : int, optional
            by default None
        manual_sort_order : int, optional
            by default None
        core_dump : bool, optional
            by default None
        """

        self.payload = {}

        if name:
            self.payload["name"] = name
        if location:
            self.payload["location"] = location
        if autostop is not None:
            self.payload["autostop"] = autostop
        if autostop_minutes is not None:
            self.payload["autostop_minutes"] = autostop_minutes
        if mysql is not None:
            self.payload["enable_mysql"] = mysql
        if reboot_on_crash is not None:
            self.payload["reboot_on_crash"] = reboot_on_crash
        if core_dump is not None:
            self.payload["enable_core_dump"] = core_dump
        if custom_domain:
            self.payload["custom_domain"] = custom_domain
        if scheduled_commands:
            self.payload["scheduled_commands"] = scheduled_commands
        if user_data:
            self.payload["user_data"] = user_data
        if manual_sort_order is not None:
            self.payload["manual_sort_order"] = manual_sort_order
        if max_disk_usage_gb is not None:
            if max_disk_usage_gb > 100 or max_disk_usage_gb < 30:
                raise InvalidStorageSize()

            self.payload["max_disk_usage_gb"] = max_disk_usage_gb

    def csgo(self, slots: int = None, tickrate: int = None,
             game_token: str = None, rcon_password: str = None,
             game_mode: str = COMPETITIVE, autoload_configs: List[str] = None,
             disable_bots: bool = False, workshop_start_map_id: int = None,
             csay_plugin: bool = False, gotv: bool = False,
             sourcemod: bool = False, insecure: bool = False,
             map_group: str = MAP_GROUP, start_map: str = None,
             password: str = None, pure: bool = True,
             admins: List[Any] = None, plugins: List[Any] = None,
             steam_key: str = None,
             workshop_id: int = None, maps_source: str = None
             ) -> ServerSettings:
        """Used for configuring a CS: GO server.

        Parameters
        ----------
        slots : int
        game_token : str
        tickrate : int
        game_mode : str, optional
            by default COMPETITIVE
        autoload_configs : List[str], optional
            by default None
        disable_bots : bool, optional
            by default False
        csay_plugin : bool, optional
            by default False
        gotv : bool, optional
            by default False
        sourcemod : bool, optional
            by default False
        insecure : bool, optional
            by default False
        map_group : str, optional
            by default MAP_GROUP
        start_map : str, optional
            by default None
        password : str, optional
            by default None
        pure : bool, optional
            by default True
        rcon_password : str, optional
            by default None
        admins : List[Any], optional
            by default None
        plugins : List[Any], optional
            by default None
        steam_key : str, optional
            by default None
        workshop_id : int, optional
            by default None
        workshop_start_map_id : int, optional
            by default None
        maps_source : int, optional
            by default None

        Raises
        ------
        MultipleGames
            Raised when you attempt to create one server
            with multiple games.
        InvalidSlotSize
            Raised when slot size is below 5 or above 64.
        InvalidTickrate
            Raised when tickrate is invalid.

        Returns
        -------
        ServerSettings
        """

        if self.__game:
            raise MultipleGames()

        self.__game = True

        self.payload["game"] = "csgo"

        if autoload_configs:
            self.payload["csgo_settings.autoload_configs"] = autoload_configs
        if disable_bots:
            self.payload["csgo_settings.disable_bots"] = disable_bots
        if csay_plugin:
            self.payload[
                "csgo_settings.enable_csay_plugin"
            ] = csay_plugin
        if gotv:
            self.payload["csgo_settings.enable_gotv"] = gotv
        if sourcemod:
            self.payload["csgo_settings.enable_sourcemod"] = sourcemod
        if game_mode:
            self.payload["csgo_settings.game_mode"] = game_mode
        if insecure is not None:
            self.payload["csgo_settings.insecure"] = insecure
        if map_group:
            self.payload["csgo_settings.mapgroup"] = map_group
        if start_map:
            self.payload["csgo_settings.mapgroup_start_map"] = start_map
        if password:
            self.payload["csgo_settings.password"] = password
        if pure is not None:
            self.payload["csgo_settings.pure_server"] = pure
        if rcon_password:
            self.payload["csgo_settings.rcon"] = rcon_password
        if slots:
            if slots < 5 or slots > 64:
                raise InvalidSlotSize()

            self.payload["csgo_settings.slots"] = slots
        if admins:
            self.payload["csgo_settings.sourcemod_admins"] = []
            admins_append = self.payload[
                "csgo_settings.sourcemod_admins"
            ].append

            for steam_id in admins:
                admins_append(
                    SteamID(steam_id).as_steam2_zero
                )
        if plugins:
            self.payload["csgo_settings.sourcemod_plugins"] = plugins
        if game_token:
            self.payload[
                "csgo_settings.steam_game_server_login_token"
            ] = game_token
        if tickrate is not None:
            if tickrate not in VALID_TICKRATES:
                raise InvalidTickrate()
            self.payload["csgo_settings.tickrate"] = tickrate
        if steam_key:
            self.payload["csgo_settings.workshop_authkey"] = steam_key
        if workshop_id:
            self.payload["csgo_settings.workshop_id"] = workshop_id
        if workshop_start_map_id:
            self.payload[
                "csgo_settings.workshop_start_map_id"
            ] = workshop_start_map_id
        if maps_source:
            self.payload["csgo_settings.maps_source"] = maps_source

        return self

    def mumble(self, slots: int = None, superuser_password: str = None,
               password: str = None, motd: str = None) -> ServerSettings:
        """Used for configuring a Mumble server.

        Parameters
        ----------
        slots : int
        superuser_password : str
        password : str, optional
            by default None
        motd : str, optional
            by default None

        Raises
        ------
        MultipleGames
            Raised when you attempt to create one server
            with multiple games.
        InvalidSlotSize
            Raised when slot size is below 7 or above 700.

        Returns
        -------
        ServerSettings
        """

        if self.__game:
            raise MultipleGames()

        self.__game = True

        self.payload["game"] = "mumble"

        if slots is not None:
            if slots < 7 or slots > 700:
                raise InvalidSlotSize()

            self.payload["mumble_settings.slots"] = slots
        if superuser_password:
            self.payload[
                "mumble_settings.superuser_password"
            ] = superuser_password
        if password:
            self.payload["mumble_settings.password"] = password
        if motd:
            self.payload["mumble_settings.welcome_text"] = motd

        return self

    def tf2(self, slots: int = None, rcon_password: str = None,
            gotv: bool = False, sourcemod: bool = False,
            insecure: bool = False, password: str = None,
            admins: list = None) -> ServerSettings:
        """Used for configuring a TF2 server.

        Parameters
        ----------
        rcon_password : str
        slots : int
        gotv : bool, optional
            by default False
        sourcemod : bool, optional
            by default False
        insecure : bool, optional
            by default False
        password : str, optional
            by default None
        admins : list, optional
            by default None

        Raises
        ------
        MultipleGames
            Raised when you attempt to create one server
            with multiple games.
        InvalidSlotSize
            Raised when slot size is below 5 or above 32.

        Returns
        -------
        ServerSettings
        """

        if self.__game:
            raise MultipleGames()

        self.__game = True

        self.payload["game"] = "teamfortress2"

        if slots is not None:
            if slots < 5 or slots > 32:
                raise InvalidSlotSize()

            self.payload["teamfortress2_settings.slots"] = slots
        if rcon_password:
            self.payload["teamfortress2_settings.rcon"] = rcon_password
        if gotv is not None:
            self.payload["teamfortress2_settings.enable_gotv"] = gotv
        if sourcemod is not None:
            self.payload[
                "teamfortress2_settings.enable_sourcemod"
            ] = sourcemod
        if insecure is not None:
            self.payload["teamfortress2_settings.insecure"] = insecure
        if password:
            self.payload["teamfortress2_settings.password"] = password
        if admins:
            self.payload["teamfortress2_settings.sourcemod_admins"] = []

            admins_append = self.payload[
                "teamfortress2_settings.sourcemod_admins"
            ].append

            for steam_id in admins:
                admins_append(
                    SteamID(steam_id).as_steam2_zero
                )

        return self

    def valheim(self, password: str = None, world_name: str = None,
                plus: bool = None, admins: List[Any] = None
                ) -> ServerSettings:
        """Used to configure valheim server.

        Parameters
        ----------
        password : str, optional
            by default None
        world_name : str, optional
            by default None
        plus : bool, optional
            by default None
        admins : List[Any], optional
            List of SteamIDs in any format, by default None

        Returns
        -------
        ServerSettings

        Raises
        ------
        MultipleGames
        """

        if self.__game:
            raise MultipleGames()

        self.payload["game"] = "valheim"
        if password:
            self.payload["valheim_settings.password"] = password
        if world_name:
            self.payload["valheim_settings.world_name"] = world_name
        if plus is not None:
            self.payload["valheim_settings.enable_valheimplus"] = plus
        if admins:
            self.payload["valheim_settings.admins_steamid64"] = []

            admins_append = self.payload[
                "valheim_settings.admins_steamid64"
            ].append

            for steam_id in admins:
                admins_append(
                    SteamID(steam_id).as_64
                )

        return self

    def teamspeak(self, slots: int) -> ServerSettings:
        """Used for configuring a teamspeak server.

        Parameters
        ----------
        slots : int

        Raises
        ------
        MultipleGames
            Raised when you attempt to create one server
            with multiple games.
        InvalidSlotSize
            Raised when slot size is below 5 or above 500.

        Returns
        -------
        ServerSettings
        """

        if self.__game:
            raise MultipleGames()

        self.__game = True

        if slots < 5 or slots > 500:
            raise InvalidSlotSize()

        self.payload["game"] = "teamspeak3"
        self.payload["teamspeak3_settings.slots"] = slots

        return self


class MatchSettings:
    def __init__(self, connection_time: int = 300,
                 knife_round: bool = False,
                 wait_for_spectators: bool = True,
                 warmup_time: int = 15) -> None:
        """Used to create a match.

        Parameters
        ----------
        connection_time : int, optional
            by default 300
        knife_round : bool, optional
            by default False
        wait_for_spectators : bool, optional
            by default True
        warmup_time : int, optional
            by default 15
        """

        self.payload = {
            "connection_time": connection_time,
            "enable_knife_round": knife_round,
            "wait_for_spectators": wait_for_spectators,
            "warmup_time": warmup_time
        }

    def __convert_id(self, given_id: Any) -> str:
        """Converts any steamID format to 32.

        Parameters
        ----------
        given_id : Any
            Given steamID.

        Returns
        -------
        str
            SteamID32

        Raises
        ------
        InvalidSteamID
            Raised when the given ID isn't understood.
        """

        steam_id = SteamID(given_id)
        if not steam_id.is_valid():
            raise InvalidSteamID()

        return steam_id.as_steam2

    def __format_players(self, players: list) -> None:
        return ",".join([self.__convert_id(steam_id) for steam_id in players])

    def playwin(self, webhook: str = None) -> MatchSettings:
        """Enables playwin AC.

        Parameters
        ----------
        webhook : str, optional
            Webhook to push playwin results, by default None

        Returns
        -------
        MatchSettings
        """

        self.payload["enable_playwin"] = True
        if webhook:
            self.payload["playwin_result_webhook_url"] = webhook

        return self

    def webhook(self, match_end: str, round_end: str,
                authorization: str = None) -> MatchSettings:
        """Used to set webhooks.

        Parameters
        ----------
        match_end : str
            URL of match end webhook.
        round_end : str
            URL of round end webhook.
        authorization : str, optional
            by default None

        Returns
        -------
        MatchSettings
        """

        self.payload["match_end_webhook_url"] = match_end
        self.payload["round_end_webhook_url"] = round_end

        if authorization:
            self.payload["webhook_authorization_header"] = authorization

        return self

    def spectators(self, players: list) -> MatchSettings:
        """Spectators

        Parameters
        ----------
        players : list
            List of spectator steam IDs,
            steamID 64, 32 & u are supported.

        Returns
        -------
        MatchSettings
        """

        self.payload["spectator_steam_ids"] = self.__format_players(players)

        return self

    def team_1(self, players: list) -> MatchSettings:
        """Team 1 players

        Parameters
        ----------
        players : list
            List of spectator steam IDs,
            steamID 64, 32 & u are supported.

        Returns
        -------
        MatchSettings
        """

        self.payload["team1_steam_ids"] = self.__format_players(players)

        return self

    def team_2(self, players: list) -> MatchSettings:
        """Team 2 players

        Parameters
        ----------
        players : list
            List of spectator steam IDs,
            steamID 64, 32 & u are supported.

        Returns
        -------
        MatchSettings
        """

        self.payload["team2_steam_ids"] = self.__format_players(players)

        return self



from datetime import datetime
from typing import Generator


class PortsModel:
    """Holds details on ports

    Attributes
    ----------
    game : int
    gotv : int
    """

    def __init__(self, data: dict) -> None:
        self.game = data["game"]
        self.gotv = data["gotv"]


class ScheduledCommandsModel:
    """Holds details on scheduled commands

    Attributes
    ----------
    name : str
    action : str
    command : str
    run_at : str
    repeat : bool
    """

    def __init__(self, data: dict) -> None:
        self.name = data["name"]
        self.action = data["action"]
        self.command = data["command"]
        self.run_at = data["run_at"]
        self.repeat = data["repeat"]


class TeamspeakModel:
    """Holds details on teamspeak server

    Attributes
    ----------
    slots : int
    admin_token : str
    """

    def __init__(self, data: dict) -> None:
        self.slots = data["slots"]
        self.admin_token = data["ts_admin_token"]


class TeamFortressModel:
    """Holds details on tf2.

    Attributes
    ----------
    slots : int
    rcon : str
    password : str
    admins : str
    gotv : bool
    sourcemod : bool
    insecure : bool
    """

    def __init__(self, data: dict) -> None:
        self.slots = data["slots"]
        self.rcon = data["rcon"]
        self.password = data["password"]
        self.admins = data["sourcemod_admins"]
        self.gotv = data["enable_gotv"]
        self.sourcemod = data["enable_sourcemod"]
        self.insecure = data["insecure"]


class MumbleModel:
    """Holds details on mumble server

    Attributes
    ----------
    slot : int
    password : str
    superuser_password : str
    motd : str
    """

    def __init__(self, data: dict) -> None:
        self.slots = data["slots"]
        self.password = data["password"]
        self.superuser_password = data["superuser_password"]
        self.motd = data["welcome_text"]


class CsgoModel:
    """Holds details on csgo server

    Attributes
    ----------
    slots : int
    game_token : str
    rcon : str
    password : str
    maps_source : str
    map_group : str
    map_group_start_map : str
    workshop_id : str
    workshop_start_map_id : str
    steam_key : str
    autoload_configs : list
    admins : str
    plugins : list
    gotv : bool
    sourcemod : bool
    csay_plugin : bool
    game_mode : str
    tickrate : int
    pure : bool
    insecure : bool
    disable_bots: bool
    """

    def __init__(self, data: dict) -> None:
        self.slots = data["slots"]
        self.game_token = data["steam_game_server_login_token"]
        self.rcon = data["rcon"]
        self.password = data["password"]
        self.maps_source = data["maps_source"]
        self.map_group = data["mapgroup"]
        self.map_group_start_map = data["mapgroup_start_map"]
        self.workshop_id = data["workshop_id"]
        self.workshop_start_map_id = data["workshop_start_map_id"]
        self.steam_key = data["workshop_authkey"]
        self.autoload_configs = data["autoload_configs"]
        self.admins = data["sourcemod_admins"]
        self.plugins = data["sourcemod_plugins"]
        self.gotv = data["enable_gotv"]
        self.sourcemod = data["enable_sourcemod"]
        self.csay_plugin = data["enable_csay_plugin"]
        self.game_mode = data["game_mode"]
        self.tickrate = data["tickrate"]
        self.pure = data["pure_server"]
        self.insecure = data["insecure"]
        self.disable_bots = data["disable_bots"]


class ServerModel:
    """Holds details on server

    Attributes
    ----------
    server_id : str
    name : str
    user_data : str
    match_id : str
    game : str
    location : str
    players_online : int
    status : list
    booting : bool
    server_error : str
    ip : str
    raw_ip : str
    on : bool
    ports : PortsModel
    confirmed : bool
    cost_per_hour : int
    max_cost_per_hour : int
    month_credits : float
    month_reset_at : datetime
    max_cost_per_month : float
    mysql : bool
    autostop : bool
    autostop_minutes : int
    mysql_username : str
    mysql_password : str
    ftp_password : str
    disk_usage_bytes : int
    default_file_locations : list
    custom_domain :  str
    added_voice_server : str
    duplicate_source_server : str
    teamspeak : TeamspeakModel
    teamfortress : TeamFortressModel
    mumble : MumbleModel
    csgo : CsgoModel
    max_disk_usage_gb : int
    reboot_on_crash : bool
    core_dump : bool
    """

    def __init__(self, data: dict) -> None:
        self.server_id = data["id"]
        self.name = data["name"]
        self.user_data = data["user_data"]
        self.match_id = data["match_id"]
        self.game = data["game"]
        self.location = data["location"]
        self.players_online = data["players_online"]
        self.status = data["status"]
        self.booting = data["booting"]
        self.server_error = data["server_error"]
        self.ip = data["ip"]
        self.raw_ip = data["raw_ip"]
        self.on = data["on"]
        self.ports = PortsModel(data["ports"])
        self.confirmed = data["confirmed"]
        self.reboot_on_crash = data["reboot_on_crash"]
        self.max_disk_usage_gb = data["max_disk_usage_gb"]
        self.core_dump = data["enable_core_dump"]
        self.cost_per_hour = data["cost_per_hour"]
        self.max_cost_per_hour = data["max_cost_per_hour"]
        self.month_credits = data["month_credits"]
        self.month_reset_at = datetime.fromtimestamp(data["month_reset_at"])
        self.max_cost_per_month = data["max_cost_per_month"]
        self.mysql = data["enable_mysql"]
        self.autostop = data["autostop"]
        self.autostop_minutes = data["autostop_minutes"]
        self.mysql_username = data["mysql_username"]
        self.mysql_password = data["mysql_password"]
        self.ftp_password = data["ftp_password"]
        self.disk_usage_bytes = data["disk_usage_bytes"]
        self.default_file_locations = data["default_file_locations"]
        self.custom_domain = data["custom_domain"]
        self.added_voice_server = data["added_voice_server"]
        self.duplicate_source_server = data["duplicate_source_server"]

        self.teamspeak = TeamspeakModel(
            data["teamspeak3_settings"]
        ) if data["teamspeak3_settings"] else None
        self.teamfortress = TeamFortressModel(
            data["teamfortress2_settings"]
        ) if data["teamfortress2_settings"] else None
        self.mumble = MumbleModel(
            data["mumble_settings"]
        ) if data["mumble_settings"] else None
        self.csgo = CsgoModel(
            data["csgo_settings"]
        ) if data["csgo_settings"] else None

        self.__scheduled_commands = data["scheduled_commands"]

    def scheduled_commands(self
                           ) -> Generator[ScheduledCommandsModel, None, None]:
        """Lists scheduled commands.

        Yields
        ------
        ScheduledCommandsModel
            Holds data on scheduled commands.
        """

        for data in self.__scheduled_commands:
            yield ScheduledCommandsModel(data)



from ..base import BackupBase

from ...routes import SERVER


class AwaitingBackup(BackupBase):
    async def restore(self) -> None:
        """Used to restore a backup.
        """

        await self.context._post(
            url=SERVER.backup_restore.format(self.server_id, self.backup_name),
        )



from ...routes import SERVER

from ..base import FileBase

from ...exceptions import AwaitingOnly


class BlockingFile(FileBase):
    def delete(self) -> None:
        """Deletes file.
        """

        self.context._delete(
            SERVER.file_interact.format(self.server_id, self.file_path)
        )

    def move(self, destination: str) -> None:
        """Used for moving a file.

        Parameters
        ----------
        destination : str

        Notes
        ------
        When called the file_path changes to the given destination.
        """

        self.context._put(
            SERVER.file_interact.format(self.server_id, self.file_path),
            data={
                "destination": destination,
            }
        )

        self.file_path = destination

    def unzip(self, destination: str) -> None:
        """Used to unzip a file.

        Parameters
        ----------
        destination : str
        """

        self.context._post(
            url=SERVER.file_unzip.format(self.server_id, self.file_path),
            data={
                "destination": destination,
            }
        )

    def upload_file(self, local_pathway: str) -> None:
        """Used to upload a local file.

        Parameters
        ----------
        local_pathway : str
            Local file to upload.
        """

        with open(local_pathway, "rb") as f:
            self.upload(f.read())

    def upload(self, data: bytes) -> None:
        """Used for uploading raw bytes.

        Parameters
        ----------
        data : bytes
            Data to upload.
        """

        self.context._post(
            url=SERVER._upload.format(self.server_id, self.file_path),
            files={
                "file": data,
            },
        )

    def save(self, local_pathway: str) -> None:
        """Saves file to local pathway.

        Parameters
        ----------
        local_pathway : str
            Pathway to save file to.
        """

        with open(local_pathway, "wb+") as f:
            f.write(self.dowload())

    def download_iterate(self) -> None:
        """
        Raises
        ------
        AwaitingOnly
            This function is meant only for awaiting
            code.
        """

        # This only exists so API structure remains identical.
        raise AwaitingOnly()

    def dowload(self) -> bytes:
        """Used to download a file into memory.

        Returns
        -------
        bytes

        Notes
        -----
        Its reccomened to use download_iterate for large files.
        """

        return self.context._get(
            SERVER.file_interact.format(self.server_id, self.file_path),
            read_bytes=True,
            read_json=False
        )



import typing

from ..base import ServerBase

from ...models.server import ServerModel
from ...models.file import FileModel
from ...models.backup import BackupModel
from ...models.metrics import MetricsModel
from ...models.match import MatchModel

from ...match.awaiting import AwaitingMatch

from .backup import AwaitingBackup
from .file import AwaitingFile

from ...settings import ServerSettings, MatchSettings

from ...exceptions import InvalidConsoleLine

from ...routes import SERVER, MATCHES


class ServerAwaiting(ServerBase):
    async def create_match(self, match_settings: MatchSettings,
                           ) -> typing.Tuple[MatchModel, AwaitingMatch]:
        """Creates a match.

        Parameters
        ----------
        match_settings : MatchSettings
            Holds details on the match.

        Returns
        -------
        MatchModel
            Holds match details.
        AwaitingMatch
            Used to interact with a match.
        """

        data = await self.context._post(
            MATCHES.create,
            data={"game_server_id": self.server_id, **match_settings.payload},
            read_json=True
        )

        return MatchModel(data), AwaitingMatch(self.context, data["id"])

    async def delete(self) -> None:
        """Used to delete a sever.
        """

        await self.context._delete(
            SERVER.delete.format(self.server_id),
        )

    async def get(self) -> ServerModel:
        """Used to get details on server.

        Returns
        -------
        ServerModel
            Holds data on server.
        """

        return ServerModel(
            await self.context._get(SERVER.get.format(self.server_id))
        )

    async def update(self, settings: ServerSettings) -> None:
        """Update servers paramters.

        Parameters
        ----------
        settings : ServerSettings
            Used to configure server.
        """

        await self.context._put(
            SERVER.update.format(self.server_id),
            data={
                **settings.payload,
                "server_id": self.server_id
            }
        )

    async def console_send(self, line: str) -> None:
        """Used to send a rcon command to console.

        Parameters
        ----------
        line : str
            Console command.
        """

        await self.context._post(
            url=SERVER.console.format(self.server_id),
            data={
                "line": line,
            }
        )

    async def console_retrive(self, lines: int = 1000) -> list:
        """Used to retrive lines from the console.

        Parameters
        ----------
        lines : int, optional
            Amount of lines to retrive, by default 1000

        Returns
        -------
        list
            List of strings.

        Raises
        ------
        InvalidConsoleLine
            Raised when console lines below 1 or above 100000.
        """

        if lines < 1 or lines > 100000:
            raise InvalidConsoleLine()

        data = await self.context._get(
            url=SERVER.console.format(self.server_id),
            params={
                "max_lines": lines,
            },
        )

        return data["lines"]

    async def sync(self) -> None:
        """Used to sync files from server to cache.
        """

        await self.context._post(
            url=SERVER.sync.format(self.server_id)
        )

    async def duplicate(self, sync: bool = False,
                        ) -> typing.Tuple[ServerModel, ServerBase]:
        """Used to duplicate a server.

        Parameters
        ----------
        sync : bool
            Used to force update server cache, by default False

        Returns
        -------
        ServerModel
            Holds server data.
        ServerAwaiting
            Used to interact with server.
        """

        if sync:
            await self.sync()

        data = await self.context._post(
            url=SERVER.duplicate.format(self.server_id),
            read_json=True,
        )

        return ServerModel(data), ServerAwaiting(self.context, data["id"])

    async def ftp_reset(self) -> None:
        """Resets the FRP password.
        """

        await self.context._post(
            url=SERVER.ftp.format(self.server_id)
        )

    async def stop(self) -> None:
        """Used to stop the server.
        """

        await self.context._post(
            url=SERVER.stop.format(self.server_id),
        )

    async def start(self, allow_host_reassignment: bool = True) -> None:
        """Used to start the server.

        Parameters
        ----------
        allow_host_reassignment : bool, optional
            By default True
        """

        await self.context._post(
            url=SERVER.start.format(self.server_id),
            data={"allow_host_reassignment": allow_host_reassignment}
        )

    async def reset(self) -> None:
        """Used to restart the server.
        """

        await self.context._post(
            url=SERVER.reset.format(self.server_id),
        )

    async def files(self, hide_default: bool = False, path: str = None,
                    file_sizes: bool = False
                    ) -> typing.AsyncGenerator[FileModel, None]:
        """Used to list files.

        Parameters
        ----------
        hide_default : bool, optional
            by default False
        path : str, optional
            Path to use as root, by default None
        file_sizes : bool, optional
            by default False

        Yields
        ------
        FileModel
            Holds details on a file.
        """

        data = await self.context._get(
            SERVER.files.format(self.server_id),
            params={
                "hide_default_files": hide_default,
                "path": path,
                "with_filesizes": file_sizes,
            },
        )

        for file_ in data:
            yield FileModel(file_), self.file(file_["path"])

    def file(self, pathway: str) -> AwaitingFile:
        """Used to interact with a file on the server.

        Parameters
        ----------
        pathway : str
            Pathway of file on server.

        Returns
        -------
        AwaitingFile
        """

        return AwaitingFile(
            self.context,
            self.server_id,
            pathway
        )

    async def backups(self
                      ) -> typing.AsyncGenerator[BackupModel, AwaitingBackup]:
        """Used to list backups a server has.

        Yields
        -------
        BackupModel
            Holds details on backup.
        AwaitingBackup
            Used for interacting with a backup.
        """

        data = await self.context._get(
            SERVER.backups.format(self.server_id),
        )

        for backup in data:
            yield BackupModel(backup), self.backup(backup["name"])

    def backup(self, backup_name: str) -> AwaitingBackup:
        """Used to interact with a backup.

        Parameters
        ----------
        backup_name : str
            Name of backup.

        Returns
        -------
        AwaitingBackup
        """

        return AwaitingBackup(
            self.context,
            self.server_id,
            backup_name
        )

    async def metrics(self) -> MetricsModel:
        """Used to get server metrics.

        Returns
        -------
        MetricsModel
            Holds details on server metrics.
        """

        data = await self.context._get(
            SERVER.metrics.format(self.server_id)
        )

        return MetricsModel(data)



class AccountModel:
    """Holds infomation around a account.

    Attributes
    ----------
    account_id : str
    email : str
    gravatar_url : str
    credits : str
    seconds_left : int
    time_left : int
    roles : str
    trial : bool
    accepted_terms_of_service_version : int
    """

    def __init__(self, data: dict) -> None:
        self.account_id = data["id"]
        self.email = data["email"]
        self.gravatar_url = data["gravatar_url"]
        self.credits = data["credits"]
        self.seconds_left = data["seconds_left"]
        self.time_left = data["time_left"]
        self.roles = data["roles"]
        self.trial = data["trial"]
        self.accepted_terms_of_service_version = (
            data["accepted_terms_of_service_version"]
        )



import typing

from .base import ServerBase

from ..models.server import ServerModel
from ..models.file import FileModel
from ..models.backup import BackupModel

from .backup import Backup

from ..settings import ServerSettings

from ..exceptions import InvalidConsoleLine

from ..routes import SERVER


class ServerBlocking(ServerBase):
    def delete(self) -> None:
        """Used to delete a sever.
        """

        self.context._delete(
            SERVER.delete.format(self.server_id)
        )

    def get(self) -> ServerModel:
        """Used to get details on server.

        Returns
        -------
        ServerModel
            Holds data on server.
        """

        return ServerModel(
            self.context._get(SERVER.get.format(self.server_id))
        )

    def update(self, settings: ServerSettings) -> None:
        """Update servers paramters.

        Parameters
        ----------
        settings : ServerSettings
            Used to configure server.
        """

        self.context._put(
            SERVER.update.format(self.server_id),
            data={
                **settings.playload,
                "server_id": self.server_id
            }
        )

    def console_send(self, line: str) -> None:
        """Used to send a command to console.

        Parameters
        ----------
        line : str
            Console command.
        """

        self.context._post(
            url=SERVER.console.format(self.server_id),
            data={
                "line": line,
            }
        )

    def console_retrive(self, lines: int = 1000) -> list:
        """Used to retrive lines from the console.

        Parameters
        ----------
        lines : int, optional
            Amount of lines to retrive, by default 1000

        Returns
        -------
        list
            List of strings.

        Raises
        ------
        InvalidConsoleLine
            Raised when console lines below 1 or above 100000.
        """
        if lines < 1 or lines > 100000:
            raise InvalidConsoleLine()

        data = self.context._get(
            url=SERVER.console.format(self.server_id),
            params={
                "max_lines": lines,
            },
        )

        return data["lines"]

    def sync(self) -> None:
        """Used to sync files from server to cache.
        """

        self.context._post(
            url=SERVER.sync.format(self.server_id)
        )

    def duplicate(self, sync: bool = False) -> (ServerModel, ServerBase):
        """Used to duplicate a server.

        Parameters
        ----------
        sync : bool
            Used to force update server cache, by default False

        Returns
        -------
        ServerModel
            Holds server data.
        ServerBlocking
            Used to interact with server.
        """

        if sync:
            self.sync()

        data = self.context._post(
            url=SERVER.duplicate.format(self.server_id),
            read_json=True
        )

        return ServerModel(data), ServerBlocking(self.context, data["id"])

    def ftp_reset(self) -> None:
        """Resets the FRP password.
        """

        self.context._post(
            url=SERVER.ftp.format(self.server_id)
        )

    def stop(self, timeout: int = 60) -> None:
        """Used to stop the server.

        Parameters
        ----------
        timeout : int, optional
            by default 60
        """

        self.context._post(
            url=SERVER.stop.format(self.server_id),
            timeout=timeout
        )

    def start(self, timeout: int = 60) -> None:
        """Used to stop the server.

        Parameters
        ----------
        timeout : int, optional
            by default 60
        """

        self.context._post(
            url=SERVER.start.format(self.server_id),
            timeout=timeout
        )

    def reset(self, timeout: int = 60) -> None:
        """Used to restart the server.

        Parameters
        ----------
        timeout : int, optional
            by default 60
        """

        self.context._post(
            url=SERVER.reset.format(self.server_id),
            timeout=timeout
        )

    def files(self, hide_default: bool = False, path: str = None,
              file_sizes: bool = False, timeout: int = 30
              ) -> typing.AsyncGenerator[FileModel, None]:
        """Used to list files.

        Parameters
        ----------
        hide_default : bool, optional
            by default False
        path : str, optional
            Path to use as root, by default None
        file_sizes : bool, optional
            by default False
        timeout : int
            by default 30

        Yields
        ------
        FileModel
            Holds details on a file.
        """

        data = self.context._get(
            SERVER.files.format(self.server_id),
            params={
                "hide_default_files": hide_default,
                "path": path,
                "with_filesizes": file_sizes,
            },
            timeout=timeout
        )

        for file_ in data:
            yield FileModel(file_)

    def backups(self, timeout: int = 30
                ) -> typing.AsyncGenerator[BackupModel, Backup]:
        """Used to list backups a server has.

        Parameters
        ----------
        timeout : int, optional
            by default 30

        Yields
        -------
        BackupModel
            Holds details on backup.
        Backup
            Used for interacting with a backup.
        """

        data = self.context._get(
            SERVER.backups.format(self.server_id),
            timeout=timeout
        )

        for backup in data:
            yield BackupModel(backup), Backup(backup["name"])



class ServerBase:
    def __init__(self, context: object, server_id: str) -> None:
        """Used to interact with a server.

        Parameters
        ----------
        context : object
            Context of the client.
        server_id : str
            Dathost server ID.
        """

        self.context = context
        self.server_id = server_id


class FileBase:
    def __init__(self, context: object,
                 server_id: str, file_path: str) -> None:
        self.context = context
        self.server_id = server_id
        self.file_path = file_path


class BackupBase:
    def __init__(self, context: object, server_id: str,
                 backup_name: str) -> None:

        self.backup_name = backup_name
        self.context = context
        self.server_id = server_id



import typing

from .base import BaseHttp


class AwaitingHttp(BaseHttp):
    async def _get(self, url, read_bytes: bool = False,
                   read_json: bool = True, *args, **kwargs) -> dict:
        """Wrapped HTTPX Delete.
        """

        return self.handle_resp(
            await self._client.get(url, *args, **kwargs),
            read=read_bytes,
            json=read_json
        )

    async def _delete(self, url, *args, **kwargs) -> bool:
        """Wrapped HTTPX Delete.
        """

        return self.handle_resp(
            await self._client.delete(url, *args, **kwargs),
            False
        )

    async def _post(self, url, read_json: bool = False,
                    *args, **kwargs) -> bool:
        """Wrapped HTTPX Post.
        """

        return self.handle_resp(
            await self._client.post(url, *args, **kwargs),
            read_json
        )

    async def _put(self, url, *args, **kwargs) -> bool:
        """Wrapped HTTPX Put.
        """

        return self.handle_resp(
            await self._client.put(url, *args, **kwargs),
            False
        )

    async def _stream(self, url, *args, **kwargs
                      ) -> typing.AsyncGenerator[bytes, None]:
        """Wrapped HTTPX stream GET.

        Yields
        -------
        bytes
        """

        async with self._client.stream("GET", url, *args, **kwargs) as resp:
            if resp.status_code == 200:
                async for chunk in resp.aiter_bytes():
                    yield chunk



import asynctest

from secrets import token_urlsafe

from ..models.account import AccountModel
from ..models.server import ServerModel
from ..models.file import FileModel
from ..models.backup import BackupModel
from ..models.metrics import MetricsModel, MapsModel, PlayerModel, \
    PlayersOnlineGraphModel
from ..models.match import MatchModel, TeamModel, MatchPlayerModel

from ..server.awaiting.backup import AwaitingBackup
from ..server.awaiting.file import AwaitingFile

from ..match.awaiting import AwaitingMatch

from ..server.awaiting import ServerAwaiting

from .. import Awaiting

from ..settings import ServerSettings, MatchSettings

from .shared_vars import EMAIL, PASSWORD, TEST_IMAGE_DIRETORY


class TestAwaitingClient(asynctest.TestCase):
    use_default_loop = True

    async def setUp(self):
        self.client = Awaiting(
            email=EMAIL,
            password=PASSWORD,
            timeout=360
        )

    async def tearDown(self):
        await self.client.close()

    async def test_awaiting_account(self):
        account = await self.client.account()

        self.assertTrue(isinstance(account, AccountModel))

    async def test_awaiting_domains(self):
        async for domain in self.client.domains():
            self.assertTrue(type(domain) == str)

    async def test_awaiting_context(self):
        context = Awaiting(EMAIL, PASSWORD)

        async with context as client:
            await client.account()

        async with context as client:
            async for domain in client.domains():
                pass

    async def test_awaiting_list_servers(self):
        async for data, server in self.client.servers():
            self.assertIsInstance(data, ServerModel)
            self.assertIsInstance(server, ServerAwaiting)

    async def test_awaiting_server_csgo(self):
        server_data, server = await self.client.create_server(
            ServerSettings(
                name="Awaiting CS: GO server",
                location="sydney",
            ).csgo(
                slots=5,
                game_token="",
                tickrate=128,
                rcon_password=token_urlsafe()
            )
        )

        self.assertIsInstance(server_data, ServerModel)
        self.assertIsInstance(server, ServerAwaiting)

        self.assertIsInstance(await server.get(), ServerModel)

        await server.console_send("status")
        await server.console_retrive()

        await server.start()
        await server.stop()
        await server.reset()

        await server.update(
            ServerSettings(
                name="Renamed Awaiting CS: GO server",
                location="sydney"
            ).csgo(
                slots=7
            )
        )

        async for data, f in server.files():
            self.assertIsInstance(data, FileModel)
            self.assertIsInstance(f, AwaitingFile)

        async for data, f in server.files(hide_default=True, file_sizes=True):
            self.assertIsInstance(data, FileModel)
            self.assertIsInstance(f, AwaitingFile)

        async for data, backup in server.backups():
            self.assertIsInstance(data, BackupModel)
            self.assertIsInstance(backup, AwaitingBackup)

            await backup.restore()

        metrics = await server.metrics()
        self.assertIsInstance(metrics, MetricsModel)

        for map_ in metrics.maps():
            self.assertIsInstance(map_, MapsModel)

        for player in metrics.players_online():
            self.assertIsInstance(player, PlayerModel)

        for player in metrics.all_time_players():
            self.assertIsInstance(player, PlayerModel)

        for player in metrics.players_online_graph():
            self.assertIsInstance(player, PlayersOnlineGraphModel)

        await server.ftp_reset()

        test_1_file = server.file("test.txt")
        await test_1_file.upload(b"hello world")
        await test_1_file.move("cfg")

        self.assertTrue(type(await test_1_file.dowload()) == bytes)

        self.assertIsNone(await test_1_file.delete())

        test_2_file = server.file("test.jpg")
        await test_2_file.upload_file(TEST_IMAGE_DIRETORY)
        await test_2_file.save(TEST_IMAGE_DIRETORY)

        async for data in test_2_file.download_iterate():
            self.assertTrue(type(data) == bytes)

        self.assertIsNone(await test_2_file.delete())

        _, duplicate = await server.duplicate(sync=True)
        self.assertIsNone(await duplicate.delete())

        match_data, match = await server.create_match(
            MatchSettings(
            ).team_1(
                [
                    "[U:1:116962485]",
                    76561198017567105,
                    "STEAM_0:1:186064092",
                    "76561198214871321"
                ]
            ).team_2(
                [
                    "[U:1:320762620]",
                    "STEAM_1:1:83437164",
                    76561198214871324,
                    "76561198214871323"
                ]
            )
        )

        self.assertIsInstance(match_data, MatchModel)
        self.assertIsInstance(match, AwaitingMatch)

        self.assertIsInstance(await match.get(), MatchModel)

        self.assertIsInstance(match_data.team_1, TeamModel)
        self.assertIsInstance(match_data.team_2, TeamModel)

        for player in match_data.players():
            self.assertIsInstance(player, MatchPlayerModel)

        self.assertIsNone(await server.delete())

    async def test_awaiting_server_valheim(self):
        data, server = await self.client.create_server(
            ServerSettings(
                name="Blocking valheim server",
                location="sydney"
            ).valheim(
                password=token_urlsafe(8),
                world_name="Poggers",
                plus=False,
                admins=[
                    "[U:1:116962485]",
                    76561198017567105,
                    "STEAM_0:1:186064092",
                    "76561198214871321"
                ]
            )
        )

        self.assertIsInstance(data, ServerModel)
        self.assertIsInstance(server, ServerAwaiting)

        self.assertIsInstance(await server.get(), ServerModel)

        self.assertIsNone(await server.delete())

    async def test_awaiting_server_mumble(self):
        data, server = await self.client.create_server(
            ServerSettings(
                name="Blocking Mumble server",
                location="sydney"
            ).mumble(
                slots=7,
                superuser_password=token_urlsafe()
            )
        )

        self.assertIsInstance(data, ServerModel)
        self.assertIsInstance(server, ServerAwaiting)

        self.assertIsInstance(await server.get(), ServerModel)

        self.assertIsNone(await server.delete())

    async def test_awaiting_server_tf2(self):
        data, server = await self.client.create_server(
            ServerSettings(
                name="Awaiting TF2 server",
                location="sydney"
            ).tf2(
                slots=5,
                rcon_password=token_urlsafe()
            )
        )

        self.assertIsInstance(data, ServerModel)
        self.assertIsInstance(server, ServerAwaiting)

        self.assertIsInstance(await server.get(), ServerModel)

        await server.update(
            ServerSettings(
                name="Renamed Awaiting CS: GO server",
                location="sydney"
            ).csgo(
                slots=7
            )
        )

        self.assertIsNone(await server.delete())

    async def test_awaiting_server_teamspeak(self):
        data, server = await self.client.create_server(
            ServerSettings(
                name="Blocking TF2 server",
                location="sydney"
            ).teamspeak(
                slots=5,
            )
        )

        self.assertIsInstance(data, ServerModel)
        self.assertIsInstance(server, ServerAwaiting)

        self.assertIsInstance(await server.get(), ServerModel)

        self.assertIsNone(await server.delete())



import os
import re

from setuptools import setup


def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


def get_long_description():
    with open("README.md", encoding="utf8") as f:
        return f.read()


def get_variable(variable):
    with open(os.path.join("dathost", "__init__.py")) as f:
        return re.search(
            "{} = ['\"]([^'\"]+)['\"]".format(variable), f.read()
        ).group(1)


setup(
    name="dathost",
    version=get_variable("__version__"),
    url=get_variable("__url__"),
    description=get_variable("__description__"),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author=get_variable("__author__"),
    author_email=get_variable("__author_email__"),
    install_requires=get_requirements(),
    license=get_variable("__license__"),
    packages=[
        "dathost",
        "dathost.models",
        "dathost.tests",
        "dathost.http",
        "dathost.server",
        "dathost.server.awaiting",
        "dathost.server.blocking",
        "dathost.match",
    ],
    python_requires=">=3.6",
    include_package_data=True,
    zip_safe=False
)



class Route:
    _ran = False
    _prefix = None

    def __init__(self, route: str = "https://dathost.net/api/0.1") -> None:
        """Used for formatting route objects.

        Parameters
        ----------
        route : str, optional
            URL of API, by default "https://dathost.net/api/0.1"
        """

        self.route = route

    def format(self) -> None:
        """Used to format URL.
        """

        if self._ran:
            return

        routes = [
            attr for attr in dir(self.__class__)
            if not callable(getattr(self.__class__, attr))
            and not attr.startswith("__")
            and not attr.startswith("_")
        ]

        for var_name in routes:
            value = "{}{}/{}".format(
                self.route,
                "/" + self._prefix if self._prefix else "",
                getattr(
                    self,
                    var_name
                )
            )

            if value[-1:] == "/":
                value = value[:-1]

            setattr(
                self,
                var_name,
                value
            )


class Account(Route):
    details = "account"


class CustomDomains(Route):
    details = "custom-domains"


class Server(Route):
    _prefix = "game-servers"

    create = ""
    delete = "{}"
    get = "{}"
    list = ""
    update = "{}"
    console = "{}/console"
    duplicate = "{}/duplicate"
    sync = "{}/sync-files"
    ftp = "{}/regenerate-ftp-password"
    reset = "{}/reset"
    start = "{}/start"
    stop = "{}/stop"
    files = "{}/files"
    file_interact = "{}/files/{}"
    file_unzip = "{}/unzip/{}"
    backups = "{}/backups"
    backup_restore = "{}/backups/{}"
    metrics = "{}/metrics"

    # We don't want any formatting applied to this route.
    _upload = "https://upload.dathost.net/api/0.1/game-servers/{}/files/{}"


class Matches(Route):
    _prefix = "matches"
    create = ""
    details = "{}"


ACCOUNT = Account()
CUSTOM_DOMAINS = CustomDomains()
SERVER = Server()
MATCHES = Matches()

for route in [ACCOUNT, CUSTOM_DOMAINS, SERVER, MATCHES]:
    route.format()



import typing
import logging

from httpx import Response
from json import JSONDecodeError

from ..exceptions import (
    NotFound,
    BadRequest,
    ExceededStorage,
    ServerStart
)


class BaseHttp:
    def handle_resp(self, resp: Response, json: bool = True,
                    read: bool = True) -> typing.Any:
        """Handles resp response.

        Parameters
        ----------
        resp : Response
        json : bool, optional
            by default True
        read : bool, optional
            by default True
        """

        if resp.status_code == 200:
            if json:
                return resp.json()
            elif read:
                return resp.read()
            else:
                return True
        else:
            try:
                message = resp.json()
            except JSONDecodeError:
                message = None
            else:
                logging.error(message)

            if resp.status_code == 404:
                raise NotFound(message)
            elif resp.status_code == 400:
                raise BadRequest(message)
            elif resp.status_code == 507:
                raise ExceededStorage(message)
            elif resp.status_code == 500:
                raise ServerStart(message)
            else:
                resp.raise_for_status()
